import re
import types

import pandas as pd
from pandas.core.frame import DataFrame
from pandas.errors import SettingWithCopyWarning
from friendly_traceback.about_warnings import WarningInfo, get_warning_parser
from friendly_traceback.info_variables import get_object_from_name
from friendly_traceback.typing_info import CauseInfo

parser = get_warning_parser(SettingWithCopyWarning)


@parser.add
def setting_with_copy_warning(message: str, warning_data: WarningInfo) -> CauseInfo:
    statement = re.sub(r"\s+", "", warning_data.problem_statement)
    # Unfortunately, IPython/Jupyter use some temporary files which are not
    # available after the fact to obtain the statement.
    if not statement:
        source = "".join(warning_data.lines)
        statement = re.sub(r"\s+", "", source)

    pattern_direct = re.compile(r"(.*)\.loc\[(.*)\]\[(.*)\].*")
    pattern_indirect_1 = re.compile(r"(.*)\.loc\[(.*)\].*")
    pattern_indirect_2 = re.compile(r"(.*)\[(.*)\].*")
    match = re.match(pattern_direct, statement)
    match_1 = re.match(pattern_indirect_1, statement)
    match_2 = re.match(pattern_indirect_2, statement)
    if match:
        return direct_indexing(
            statement, match[1].strip(), match[2].strip(), match[3].strip()
        )
    if match_1:
        series = match_1[1].strip()
        series_obj = get_object_from_name(series, warning_data.frame)
        if isinstance(series_obj, pd.Series):
            return indirect_indexing(
                series, match_1[2].strip(), warning_data.frame, loc_2=".loc"
            )
    elif match_2:
        series = match_2[1].strip()
        series_obj = get_object_from_name(series, warning_data.frame)
        if isinstance(series_obj, pd.Series):
            return indirect_indexing(
                series, match_2[2].strip(), warning_data.frame, loc_1=".loc"
            )
    return {}


def direct_indexing(
    statement: str, dataframe_name: str, index_1: str, index_2: str
) -> CauseInfo:
    cause = (
        "You used direct chained indexing of a dataframe which made a copy\n"
        "of the original content of the dataframe. If you try to assign a value\n"
        "to that copy, the original dataframe will not be modified.\n"
        "Instead of doing a direct chained indexing\n\n"
        "    {dataframe_name}.loc[{index_1}][{index_2}] ...\n\n"
        " try:\n\n"
        "    {dataframe_name}.loc[{index_1}, {index_2}] ...\n"
    ).format(
        statement=statement,
        dataframe_name=dataframe_name,
        index_1=index_1,
        index_2=index_2,
    )
    return {"cause": cause}


def indirect_indexing(
    series_name: str,
    index: str,
    frame: types.FrameType,
    loc_1: str = "",
    loc_2: str = "",
) -> CauseInfo:
    names = find_all_dataframes(frame)
    if not names:
        return {}
    basic_cause = (
        "I suspect that you used indirect chained indexing of a dataframe.\n"
        "First, you likely created a series using something like:\n\n"
        "    {series_name} = {dataframe_name}{loc_1}[...]\n\n"
        "This made a copy of the data contained in the dataframe.\n"
        "Next, you indexed that copy\n\n"
        "    {series_name}{loc_2}[{index}]\n\n"
        "This had no effect on the original dataframe.\n"
        "If your goal is to modify the value of the original dataframe,\n"
        "try something like the following instead:\n\n"
        "    {dataframe_name}.loc[..., {index}]\n"
    )
    name = names.pop()
    if len(names) == 0:
        cause = basic_cause.format(
            series_name=series_name,
            dataframe_name=name,
            index=index,
            loc_1=loc_1,
            loc_2=loc_2,
        )
    else:
        names.add(name)
        intro = (
            "In your code, you have the following dataframes: `{df}`.\n"
            "I do not know which one is causing the problem here;\n"
            "I will use the name `{name}` as an example.\n\n"
        ).format(df=names, name=name)
        cause = intro + basic_cause.format(
            series_name=series_name,
            dataframe_name=name,
            index=index,
            loc_1=loc_1,
            loc_2=loc_2,
        )
    return {"cause": cause}


def find_all_dataframes(frame: types.FrameType) -> set:
    dataframes = {
        name for name in frame.f_locals if isinstance(frame.f_locals[name], DataFrame)
    }
    gl_dataframes = {
        name for name in frame.f_globals if isinstance(frame.f_globals[name], DataFrame)
    }
    return dataframes.union(gl_dataframes)



from friendly_traceback.info_generic import register
from pandas.errors import SettingWithCopyWarning


@register(SettingWithCopyWarning)
def setting_with_copy_warning() -> str:
    return ("Pandas occasionally emits a `SettingWithCopyWarning` when you use 'chained indexing',\n"
            "either directly or indirectly,"
             "and you then attempt to assign a value to the result.\n"
             "By 'direct chained indexing', we mean that your code contains something like:\n\n"
             "    ...[index_1][index_2] = ...\n"
             "During the first extraction using `[index_1]`, pandas found that the series to be created\n"
             "contained values of different types. It automatically created a new series\n"
             "converting all values to a common type.\n"
             "The second indexing, `[index_2]` was then done a this copy\n"
             "instead of the original dataframe.\n"
             "Thus, the assigment was not done on the original dataframe, which caused Pandas to emit this warning.\n\n"
             "An 'indirect chained indexing' essentially amount to the same problem except that\n"
             "the second indexing is not done on the same line as that which was done\n"
             "to extract the first series.\n"
             )



import re
from pandas.core.frame import DataFrame
from friendly_traceback import utils as ft_utils
from friendly_traceback.message_parser import get_parser
from friendly_traceback.info_variables import get_object_from_name

parser = get_parser(KeyError)
# The decorator parser.add will ensure that the function loc_does_not_exist
# is going to be used in an attempt to find the cause of the error before
# any similar functions from friendly_traceback would be used to do the same.
@parser.add
def loc_does_not_exist(message, traceback_data):
    # The message for a key error is not useful as it contains only
    # the unknown key.
    # Did we try to use loc?
    match = re.search(r"(.*)\.loc", traceback_data.bad_line)
    if match is None:
        # let the other registered handlers attempt to find an explanation
        return {}

    df = match[1]
    frame = traceback_data.exception_frame
    target = get_object_from_name(df, frame)
    if target is None:  # This is very unlikely to happen
        return {}

    if isinstance(target, DataFrame):
        # In the error message, the key shown is a string representation of
        # the actual key. This is the way to get the actual key.
        key = traceback_data.value.args[0]
        columns = list(target)
        if key in columns:
            # Note the use of backticks below to surround code: this is markdown notation that
            # friendly can use to add syntax coloring.
            hint = f'To retrieve a column, just use square brackets: `{df}["{key}"]`.\n'
            return {
                "cause": "You tried to use loc to retrieve a column, but it takes a row or a row selector.\n" + hint,
                "suggest": hint
            }
        else:
            rows = list(target.index.values)
            similar = ft_utils.get_similar_words(key, rows)
            if len(similar) == 1:
                hint = ("Did you mean `{name}`?\n").format(name=similar[0])
                cause = (
                    "`{name}` is a key of `{dict_}` which is similar to `{key}`.\n"
                ).format(name=similar[0], dict_=df, key=repr(key))
                return {"cause": cause, "suggest": hint}
            elif len(similar) > 1:
                hint = f"Did you mean `{similar[0]}`?\n"
                names = ", ".join(similar)
                cause = f"`{df}` has some keys similar to `{key!r}` including:\n`{names}`.\n"
                return {"cause": cause, "suggest": hint}

            rows = ft_utils.list_to_string(rows) # Remove the brackets surrounding list items
            return {
                "cause": (
                    f"You tried to retrieve an unknown row. The valid values are:\n`{rows}`.\n"
                )
            }
    return {}



__version__ = "0.1.2"

import pathlib

import pandas
from friendly_traceback import exclude_directory_from_traceback, config

# The following import will automatically add relevant parsers to
# those known by friendly_traceback
from . import specific
from . import generic

print(f"friendly_pandas version {__version__}")

# We want to focus on the code entered by the user.
# We remove anything that occurs inside pandas' library from the traceback
_pandas_init = pathlib.Path(pandas.__file__)
_pandas_dir = _pandas_init.parents[0]
exclude_directory_from_traceback(_pandas_dir)

# Disabling showing chained exceptions in normal "friendly" tracebacks
# as these likely come from code all inside pandas library.
# This will likely become the default in a future version.
config.session.include_chained_exception = False
